@@ -21,10 +21,6 @@ use std::fs;
 use std::path::{Path, PathBuf};
 
 use anyhow::Context;
-use futures_util::{FutureExt, future::{BoxFuture, TryFutureExt}};
-use http::Response;
-use hyper::{Body, Client, Uri};
-use hyper_tls::HttpsConnector;
 
 use openpgp::policy::StandardPolicy;
 use sequoia_openpgp::{
@@ -37,8 +33,8 @@ use sequoia_openpgp::{
     cert::prelude::*,
 };
 
-use super::{Result, Error};
-use super::email::EmailAddress;
+use crate::email::EmailAddress;
+use crate::Result;
 
 /// WKD variants.
 ///
@@ -109,19 +105,9 @@ impl Url {
     }
 
     /// Returns an [`url::Url`].
-    pub fn to_url<V>(&self, variant: V) -> Result<url::Url>
+    pub fn to_url<V>(&self, variant: V) -> Result<reqwest::Url>
             where V: Into<Option<Variant>> {
-        let url_string = self.build(variant);
-        let url_url = url::Url::parse(url_string.as_str())?;
-        Ok(url_url)
-    }
-
-    /// Returns an [`hyper::Uri`].
-    pub fn to_uri<V>(&self, variant: V) -> Result<Uri>
-            where V: Into<Option<Variant>> {
-        let url_string = self.build(variant);
-        let uri = url_string.as_str().parse::<Uri>()?;
-        Ok(uri)
+        Ok(reqwest::Url::parse(self.build(variant).as_str())?)
     }
 
     /// Returns a [`PathBuf`].
@@ -152,7 +138,7 @@ fn encode_local_part<S: AsRef<str>>(local_part: S) -> String {
     let _ = ctx.digest(&mut digest);
 
     // After z-base-32 encoding 20 bytes, it will be 32 bytes long.
-    zbase32::encode_full_bytes(&digest[..])
+    zbase32::encode(&digest[..])
 }
 
 
@@ -166,64 +152,35 @@ fn encode_local_part<S: AsRef<str>>(local_part: S) -> String {
 /// address.
 /// ```
 fn parse_body<S: AsRef<str>>(body: &[u8], email_address: S)
-        -> Result<Vec<Cert>> {
+        -> Result<Vec<Result<Cert>>> {
     let email_address = email_address.as_ref();
     // This will fail on the first packet that can not be parsed.
     let packets = CertParser::from_bytes(&body)?;
     // Collect only the correct packets.
-    let certs: Vec<Cert> = packets.flatten().collect();
+    let certs: Vec<Result<Cert>> = packets.collect();
     if certs.is_empty() {
-        return Err(Error::NotFound.into());
+        return Err(crate::Error::NotFound.into());
     }
 
     // Collect only the Certs that contain the email in any of their userids
-    let valid_certs: Vec<Cert> = certs.iter()
+    let valid_certs: Vec<Result<Cert>> = certs.into_iter()
         // XXX: This filter could become a Cert method, but it adds other API
         // method to maintain
-        .filter(|cert| {cert.userids()
-            .any(|uidb|
-                if let Ok(Some(a)) = uidb.userid().email() {
-                    a == email_address
-                } else { false })
-        }).cloned().collect();
+        .filter(|cert| match cert {
+            Ok(cert) => cert.userids()
+                .any(|uidb|
+                     if let Ok(Some(a)) = uidb.userid().email2() {
+                         a == email_address
+                     } else { false }),
+            Err(_) => true,
+        }).collect();
     if valid_certs.is_empty() {
-        Err(Error::EmailNotInUserids(email_address.into()).into())
+        Err(crate::Error::EmailNotInUserids(email_address.into()).into())
     } else {
         Ok(valid_certs)
     }
 }
 
-fn get_following_redirects<T>(
-    client: &hyper::client::Client<T>,
-    url: Uri,
-    depth: i32,
-) -> BoxFuture<Result<Response<Body>>>
-where
-    T: hyper::client::connect::Connect + Clone + Send + Sync + 'static,
-{
-    async move {
-        let response = client.get(url).await;
-
-        if depth < 0 {
-            return Err(anyhow::anyhow!("Too many redirects"));
-        }
-
-        if let Ok(ref resp) = response {
-            if resp.status().is_redirection() {
-                let url = resp.headers().get("Location")
-                    .and_then(|value| value.to_str().ok())
-                    .map(|value| value.parse::<Uri>());
-                if let Some(Ok(url)) = url {
-                    return get_following_redirects(client, url, depth - 1).await;
-                }
-            }
-        }
-
-        response.map_err(|err| anyhow::anyhow!(err))
-    }
-    .boxed()
-}
-
 /// Retrieves the Certs that contain userids with a given email address
 /// from a Web Key Directory URL.
 ///
@@ -250,6 +207,7 @@ where
 /// ```
 ///
 /// [draft-koch]: https://datatracker.ietf.org/doc/html/draft-koch-openpgp-webkey-service/#section-3.1
+///
 /// # Examples
 ///
 /// ```no_run
@@ -257,41 +215,45 @@ where
 /// # use sequoia_openpgp::Cert;
 /// # async fn f() -> Result<()> {
 /// let email_address = "foo@bar.baz";
-/// let certs: Vec<Cert> = wkd::get(&email_address).await?;
+/// let certs: Vec<Result<Cert>> =
+///     wkd::get(&reqwest::Client::new(), &email_address).await?;
 /// # Ok(())
 /// # }
 /// ```
 
 // XXX: Maybe the direct method should be tried on other errors too.
 // https://mailarchive.ietf.org/arch/msg/openpgp/6TxZc2dQFLKXtS0Hzmrk963EteE
-pub async fn get<S: AsRef<str>>(email_address: S) -> Result<Vec<Cert>> {
+pub async fn get<S: AsRef<str>>(c: &reqwest::Client, email_address: S)
+                                -> Result<Vec<Result<Cert>>>
+{
     let email = email_address.as_ref().to_string();
     // First, prepare URIs and client.
     let wkd_url = Url::from(&email)?;
 
-    // WKD must use TLS, so build a client for that.
-    let https = HttpsConnector::new();
-    let client = Client::builder().build::<_, hyper::Body>(https);
-    let advanced_uri = wkd_url.to_uri(Variant::Advanced)?;
-    let direct_uri = wkd_url.to_uri(Variant::Direct)?;
-
-    const REDIRECT_LIMIT: i32 = 10;
+    let advanced_uri = wkd_url.to_url(Variant::Advanced)?;
+    let direct_uri = wkd_url.to_url(Variant::Direct)?;
 
     // First, try the Advanced Method.
-    let res = get_following_redirects(&client, advanced_uri, REDIRECT_LIMIT)
+    let res = if let Ok(res) = c.get(advanced_uri).send().await {
+        Ok(res)
+    } else {
         // Fall back to the Direct Method.
-        .or_else(|_| get_following_redirects(&client, direct_uri, REDIRECT_LIMIT))
-        .await?;
-    let body = hyper::body::to_bytes(res.into_body()).await?;
+        c.get(direct_uri).send().await
+    }.map_err(Error::NotFound)?;
 
-    parse_body(&body, &email)
+    if res.status() == reqwest::StatusCode::OK {
+        let body = res.bytes().await?;
+        parse_body(&body, &email)
+    } else {
+        Err(crate::Error::NotFound.into())
+    }
 }
 
 /// Returns all e-mail addresses from certificate's User IDs matching `domain`.
 fn get_cert_domains<'a>(domain: &'a str, cert: &ValidCert<'a>) -> impl Iterator<Item = Url> + 'a
 {
     cert.userids().filter_map(move |uidb| {
-        uidb.userid().email().unwrap_or(None).and_then(|addr| {
+        uidb.userid().email2().unwrap_or(None).and_then(|addr| {
             if EmailAddress::from(&addr).ok().map(|e| e.domain == domain)
                 .unwrap_or(false)
             {
@@ -404,6 +366,14 @@ impl KeyRing {
     }
 }
 
+/// Errors for this module.
+#[derive(thiserror::Error, Debug)]
+#[non_exhaustive]
+pub enum Error {
+    /// A requested cert was not found.
+    #[error("Cert not found")]
+    NotFound(#[from] reqwest::Error),
+}
 
 #[cfg(test)]
 mod tests {
@@ -435,10 +405,10 @@ mod tests {
              stnkabub89rpcphiz4ppbxixkwyt1pic?l=test1";
         let wkd_url = Url::from("test1@example.com").unwrap();
         assert_eq!(expected_url, wkd_url.to_string());
-        assert_eq!(url::Url::parse(expected_url).unwrap(),
+        assert_eq!(reqwest::Url::parse(expected_url).unwrap(),
+                   wkd_url.to_url(None).unwrap());
+        assert_eq!(expected_url.parse::<reqwest::Url>().unwrap(),
                    wkd_url.to_url(None).unwrap());
-        assert_eq!(expected_url.parse::<Uri>().unwrap(),
-                   wkd_url.to_uri(None).unwrap());
 
         // Direct method
         let expected_url =
@@ -446,10 +416,10 @@ mod tests {
              .well-known/openpgpkey/hu/\
              stnkabub89rpcphiz4ppbxixkwyt1pic?l=test1";
         assert_eq!(expected_url, wkd_url.build(Direct));
-        assert_eq!(url::Url::parse(expected_url).unwrap(),
+        assert_eq!(reqwest::Url::parse(expected_url).unwrap(),
+                   wkd_url.to_url(Direct).unwrap());
+        assert_eq!(expected_url.parse::<reqwest::Url>().unwrap(),
                    wkd_url.to_url(Direct).unwrap());
-        assert_eq!(expected_url.parse::<Uri>().unwrap(),
-                   wkd_url.to_uri(Direct).unwrap());
     }
 
     #[test]
